"use strict";

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

Object.defineProperty(exports, "__esModule", { value: true });
var core_1 = require("../core");
var isOrIncludes = function isOrIncludes(value, elem) {
    return Array.isArray(value) ? value.includes(elem) : value === elem;
};
function readLayer(_ref, rootRecords, state, firstIds, rootSpans) {
    var root = _ref.root,
        field = _ref.field,
        args = _ref.args,
        structuralFields = _ref.structuralFields,
        scalarFields = _ref.scalarFields,
        relations = _ref.relations,
        path = _ref.path;

    var filter = function filter(id) {
        return core_1.runFilter(args.filter, id, state.combined[field.type][id]);
    };
    var compare = core_1.createCompare(function (id, key) {
        return key === 'id' ? id : state.combined[field.type][id][key];
    }, args.sort);
    var compareRecords = core_1.createCompare(function (record, key) {
        return record[key];
    }, args.sort);
    var rootIds = Object.keys(rootRecords);
    var rootRecordIds = {};
    var sliceStarts = {};
    var records = {};
    var getRecord = function getRecord(id) {
        if (!id) return null;
        if (records[id]) return records[id];
        return records[id] = core_1.keysToObject(Object.keys(scalarFields), function (f) {
            return f === 'id' ? id : core_1.noUndef(state.combined[field.type][id][f]);
        });
    };
    var findRecordIndex = function findRecordIndex(rootId, record) {
        return core_1.locationOf('', rootRecordIds[rootId], core_1.createCompare(function (id, key) {
            return key === 'id' ? id || record.id : id ? state.combined[field.type][id][key] : record[key];
        }, args.sort));
    };
    var allIds = Object.keys(state.combined[field.type] || {});
    var filteredIdsObj = core_1.keysToObject(allIds, filter);
    var filteredIds = allIds.filter(function (id) {
        return filteredIdsObj[id];
    }).sort(compare);
    var initRootRecords = function initRootRecords(rootId) {
        if (!root.type) {
            rootRecordIds[rootId] = filteredIds;
        } else {
            var value = state.combined[root.type][rootId][root.field];
            if (core_1.fieldIs.relation(field)) {
                if (field.isList) {
                    if (!args.sort) {
                        rootRecordIds[rootId] = (value || []).map(function (id) {
                            return filteredIds.includes(id) ? id : null;
                        });
                    } else {
                        rootRecordIds[rootId] = filteredIds.filter(function (id) {
                            return (value || []).includes(id);
                        });
                    }
                } else {
                    rootRecordIds[rootId] = value && filteredIds.includes(value) ? [value] : [];
                }
            } else {
                rootRecordIds[rootId] = filteredIds.filter(function (id) {
                    return (value || []).includes(id) || isOrIncludes(state.combined[field.type][id][field.foreign], rootId);
                });
            }
        }
        if (rootRecordIds[rootId].length === 0) {
            rootRecords[rootId][root.field] = core_1.fieldIs.foreignRelation(field) || field.isList ? [] : null;
        } else if (core_1.fieldIs.relation(field) && field.isList && !args.sort) {
            rootRecords[rootId][root.field] = rootRecordIds[rootId].map(getRecord);
        } else if (core_1.fieldIs.foreignRelation(field) || field.isList) {
            var queryFirst = Object.assign({ id: firstIds[path][rootId] }, state.server[field.type][firstIds[path][rootId]]);
            var queryStart = findRecordIndex(rootId, queryFirst);
            sliceStarts[rootId] = queryStart;
            var _iteratorNormalCompletion = true;
            var _didIteratorError = false;
            var _iteratorError = undefined;

            try {
                for (var _iterator = Object.keys(state.diff[field.type] || {})[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                    var id = _step.value;

                    if (state.diff[field.type][id] === 1) {
                        var localIndex = rootRecordIds[rootId].indexOf(id);
                        if (localIndex !== -1 && localIndex < queryStart) {
                            sliceStarts[rootId] -= 1;
                        }
                    }
                    if (state.diff[field.type][id] === 0) {
                        if (state.server[field.type][id] && core_1.runFilter(args.filter, id, state.server[field.type][id]) && compareRecords(state.server[field.type][id], queryFirst) === -1) {
                            sliceStarts[rootId] += 1;
                        }
                        var _localIndex = rootRecordIds[rootId].indexOf(id);
                        if (_localIndex !== -1 && _localIndex < queryStart) {
                            sliceStarts[rootId] -= 1;
                        }
                    }
                    if (state.diff[field.type][id] === -1) {
                        var serverRecord = (state.server[field.type] || {})[id];
                        if (serverRecord && (!root.type || core_1.fieldIs.foreignRelation(field) || state.combined[root.type][rootId][root.field].includes(id)) && core_1.runFilter(args.filter, id, serverRecord)) {
                            if (compareRecords(Object.assign({ id: id }, serverRecord), queryFirst) === -1) {
                                sliceStarts[rootId] += 1;
                            }
                        }
                    }
                }
            } catch (err) {
                _didIteratorError = true;
                _iteratorError = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion && _iterator.return) {
                        _iterator.return();
                    }
                } finally {
                    if (_didIteratorError) {
                        throw _iteratorError;
                    }
                }
            }

            rootRecords[rootId][root.field] = rootRecordIds[rootId].slice(sliceStarts[rootId], core_1.undefOr(args.end, sliceStarts[rootId] + args.end - (args.start || 0))).map(getRecord);
        } else {
            rootRecords[rootId][root.field] = getRecord(rootRecordIds[rootId][0] || null);
        }
    };
    rootIds.forEach(initRootRecords);
    var recordSpans = rootSpans && core_1.keysToObject(Object.keys(records), function () {
        return {};
    });
    var relationUpdaters = relations.map(function (relationLayer) {
        return readLayer(relationLayer, records, state, firstIds, recordSpans);
    });
    if (rootSpans) {
        Object.keys(records).forEach(function (id) {
            var relationCounts = relations.map(function (_ref2) {
                var root = _ref2.root;
                return recordSpans[id][root.field].reduce(function (res, v) {
                    return res + v[''];
                }, 0);
            });
            recordSpans[id][''] = Math.max.apply(Math, _toConsumableArray(relationCounts).concat([1]));
            relations.forEach(function (_ref3, i) {
                var root = _ref3.root;

                var diff = recordSpans[id][''] - relationCounts[i];
                if (diff > 0) recordSpans[id][root.field].push(diff);
            });
        });
        rootIds.forEach(function (rootId) {
            rootSpans[rootId][root.field] = rootRecordIds[rootId].map(function (id) {
                return id ? recordSpans[id] : { '': 1 };
            });
            if (rootSpans[rootId][root.field].length === 0) {
                rootSpans[rootId][root.field] = [{ '': 1 }];
            }
        });
    }
    return function (changes, update) {
        var relationsChange = Math.max.apply(Math, _toConsumableArray(relationUpdaters.map(function (updater) {
            return updater(changes, update);
        })).concat([0]));
        if (relationsChange === 2) return 2;
        var _iteratorNormalCompletion2 = true;
        var _didIteratorError2 = false;
        var _iteratorError2 = undefined;

        try {
            for (var _iterator2 = Object.keys(changes[field.type] || {})[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                var id = _step2.value;
                var _iteratorNormalCompletion6 = true;
                var _didIteratorError6 = false;
                var _iteratorError6 = undefined;

                try {
                    for (var _iterator6 = structuralFields[Symbol.iterator](), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
                        var _f = _step6.value;

                        if ((changes[field.type][id] || {})[_f]) return 2;
                    }
                } catch (err) {
                    _didIteratorError6 = true;
                    _iteratorError6 = err;
                } finally {
                    try {
                        if (!_iteratorNormalCompletion6 && _iterator6.return) {
                            _iterator6.return();
                        }
                    } finally {
                        if (_didIteratorError6) {
                            throw _iteratorError6;
                        }
                    }
                }

                if (core_1.fieldIs.foreignRelation(field) && (changes[field.type][id] || {})[field.foreign]) {
                    return 2;
                }
            }
        } catch (err) {
            _didIteratorError2 = true;
            _iteratorError2 = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion2 && _iterator2.return) {
                    _iterator2.return();
                }
            } finally {
                if (_didIteratorError2) {
                    throw _iteratorError2;
                }
            }
        }

        if (root.type) {
            var _iteratorNormalCompletion3 = true;
            var _didIteratorError3 = false;
            var _iteratorError3 = undefined;

            try {
                for (var _iterator3 = Object.keys(changes[root.type] || {})[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
                    var _id = _step3.value;

                    if (rootRecords[_id]) {
                        if ((changes[root.type][_id] || {})[root.field]) return 2;
                    }
                }
            } catch (err) {
                _didIteratorError3 = true;
                _iteratorError3 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion3 && _iterator3.return) {
                        _iterator3.return();
                    }
                } finally {
                    if (_didIteratorError3) {
                        throw _iteratorError3;
                    }
                }
            }
        }
        var hasUpdated = false;
        if (update) {
            var _iteratorNormalCompletion4 = true;
            var _didIteratorError4 = false;
            var _iteratorError4 = undefined;

            try {
                for (var _iterator4 = Object.keys(changes[field.type] || {})[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
                    var _id2 = _step4.value;

                    if (records[_id2]) {
                        var _iteratorNormalCompletion5 = true;
                        var _didIteratorError5 = false;
                        var _iteratorError5 = undefined;

                        try {
                            for (var _iterator5 = Object.keys(changes[field.type][_id2] || {})[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
                                var f = _step5.value;

                                if (scalarFields[f]) {
                                    var prev = records[_id2][f];
                                    var value = ((state.combined[field.type] || {})[_id2] || {})[f];
                                    if (value === undefined) delete records[_id2][f];else records[_id2][f] = core_1.noUndef(value);
                                    if (records[_id2][f] !== prev) hasUpdated = true;
                                }
                            }
                        } catch (err) {
                            _didIteratorError5 = true;
                            _iteratorError5 = err;
                        } finally {
                            try {
                                if (!_iteratorNormalCompletion5 && _iterator5.return) {
                                    _iterator5.return();
                                }
                            } finally {
                                if (_didIteratorError5) {
                                    throw _iteratorError5;
                                }
                            }
                        }
                    }
                }
            } catch (err) {
                _didIteratorError4 = true;
                _iteratorError4 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion4 && _iterator4.return) {
                        _iterator4.return();
                    }
                } finally {
                    if (_didIteratorError4) {
                        throw _iteratorError4;
                    }
                }
            }
        }
        return Math.max(relationsChange, hasUpdated ? 1 : 0);
    };
    // const sliceInfo = (rootId: string, index: number) => {
    //   const end = args.show !== null ? sliceStarts[rootId] + args.show : null;
    //   return {
    //     before: index <= sliceStarts[rootId],
    //     within: index > sliceStarts[rootId] && (end === null || index < end),
    //     indexInSlice: index - sliceStarts[rootId],
    //     last: end && {
    //       index: end,
    //       id: rootRecordIds[rootId][sliceStarts[rootId] + end] as string,
    //     },
    //   };
    // };
    // const stop =
    //   onChanges &&
    //   onChanges(({ changes, rootChanges }) => {
    //     newRecords = [];
    //     const maybeRemoved: Obj<true> = {};
    //     const filteredAdded: string[] = [];
    //     const filteredRemoved: string[] = [];
    //     const foreignChanged: string[] = [];
    //     for (const id of Object.keys(changes[field.type] || {})) {
    //       filteredIdsObj[id] = filteredIdsObj[id] || false;
    //       const included = filter(id);
    //       if (included !== filteredIdsObj[id]) {
    //         if (included) {
    //           filteredAdded.push(id);
    //           const index = locationOf(id, filteredIds, compare);
    //           filteredIds.splice(index, 0, id);
    //           console.log(filteredIds);
    //           if (!root.type) {
    //             const info = sliceInfo('', index);
    //             if ((info.before || info.within) && info.last) {
    //               filteredRemoved.push(info.last.id);
    //               rootRecords[''][root.field].pop();
    //               delete records[info.last.id];
    //             }
    //             if (info.before) {
    //               rootRecords[''][root.field].unshift(
    //                 getRecord(filteredIds[sliceStarts['']]),
    //               );
    //             } else if (info.within) {
    //               rootRecords[''][root.field].splice(
    //                 info.indexInSlice,
    //                 0,
    //                 getRecord(id),
    //               );
    //             }
    //           }
    //         } else {
    //           const index = filteredIds.indexOf(id);
    //           filteredIds.splice(index, 1);
    //           if (records[id]) {
    //             filteredRemoved.push(id);
    //             delete records[id];
    //           }
    //           if (!root.type) {
    //             const info = sliceInfo('', index);
    //             if ((info.before || info.within) && info.last) {
    //               rootRecords[''][root.field].push(getRecord(info.last.id));
    //             }
    //             if (info.before) {
    //               rootRecords[''][root.field].shift();
    //             } else if (info.within) {
    //               rootRecords[''][root.field].splice(info.indexInSlice, 1);
    //             }
    //           }
    //         }
    //         filteredIdsObj[id] = !filteredIdsObj[id];
    //       } else if (
    //         included &&
    //         fieldIs.foreignRelation(field) &&
    //         ((changes[field.type] && {})[id] || {})[field.foreign]
    //       ) {
    //         foreignChanged.push(id);
    //       }
    //     }
    //     for (let i = rootIds.length - 1; i >= 0; i--) {
    //       const rootId = rootIds[i];
    //       if (rootChanges.removed.includes(rootId)) {
    //         rootRecordIds[rootId].forEach(id => id && (maybeRemoved[id] = true));
    //         rootIds.splice(i, 1);
    //         delete rootRecordIds[rootId];
    //         delete sliceStarts[rootId];
    //       } else if (
    //         root.type &&
    //         ((changes[root.type] && {})[rootId] || {})[root.field]
    //       ) {
    //         rootRecordIds[rootId].forEach(id => id && (maybeRemoved[id] = true));
    //         initRootRecords(rootId);
    //       } else {
    //         if (root.type) {
    //           const addRecord = (id: string) => {
    //             const index = locationOf(id, rootRecordIds[rootId], compare);
    //             const info = sliceInfo(rootId, index);
    //             if (info.before || info.within) {
    //               if (info.last) {
    //                 maybeRemoved[info.last.id] = true;
    //                 rootRecordIds[rootId].splice(info.last.index, 1);
    //                 rootRecords[rootId][root.field].pop();
    //               }
    //               rootRecordIds[rootId].splice(index, 0, id);
    //             }
    //             if (info.before) {
    //               rootRecords[rootId][root.field].unshift(
    //                 getRecord(filteredIds[sliceStarts['']]),
    //               );
    //             } else if (info.within) {
    //               rootRecords[rootId][root.field].splice(
    //                 info.indexInSlice,
    //                 0,
    //                 getRecord(id),
    //               );
    //             }
    //           };
    //           const removeRecord = (id: string) => {
    //             const index = rootRecordIds[rootId].indexOf(id);
    //             if (index !== -1) {
    //               const info = sliceInfo(rootId, index);
    //               if (info.before || info.within) {
    //                 if (info.last) {
    //                   rootRecordIds[rootId].splice(
    //                     info.last.index,
    //                     0,
    //                     info.last.id,
    //                   );
    //                   rootRecords[rootId][root.field].push(
    //                     getRecord(info.last.id),
    //                   );
    //                 }
    //                 rootRecordIds[rootId].splice(index, 1);
    //               }
    //               if (info.before) {
    //                 rootRecords[rootId][root.field].shift();
    //               } else if (info.within) {
    //                 rootRecords[rootId][root.field].splice(info.indexInSlice, 1);
    //               }
    //             }
    //           };
    //           const value = state.combined[root.type!][rootId]![root.field];
    //           filteredAdded.forEach(id => {
    //             if (fieldIs.relation(field)) {
    //               if (field.isList) {
    //                 if (args.unsorted) {
    //                   const index = ((value || []) as string[]).indexOf(id);
    //                   if (index !== -1) {
    //                     rootRecordIds[rootId][index] = id;
    //                     const i = index - sliceStarts[rootId];
    //                     if (i >= 0 && (args.show === null || i < args.show)) {
    //                       rootRecords[rootId][root.field][i] = getRecord(id);
    //                     }
    //                   }
    //                 } else {
    //                   if ((value || []).includes(id)) addRecord(id);
    //                 }
    //               } else {
    //                 if (value === id) {
    //                   rootRecordIds[rootId] = [id];
    //                   rootRecords[rootId][root.field] = getRecord(id);
    //                 }
    //               }
    //             } else {
    //               if (
    //                 (value || []).includes(id) ||
    //                 isOrIncludes(
    //                   state.combined[root.type!][id]![field.foreign],
    //                   rootId,
    //                 )
    //               ) {
    //                 addRecord(id);
    //               }
    //             }
    //           });
    //           filteredRemoved.forEach(id => {
    //             if (fieldIs.relation(field)) {
    //               if (field.isList) {
    //                 removeRecord(id);
    //               } else {
    //                 if (rootRecordIds[rootId][0] === id) {
    //                   rootRecordIds[rootId] = [];
    //                   rootRecords[rootId][root.field] = null;
    //                 }
    //               }
    //             } else {
    //               removeRecord(id);
    //             }
    //           });
    //           if (fieldIs.foreignRelation(field)) {
    //             foreignChanged.forEach(id => {
    //               const included =
    //                 (value || []).includes(id) ||
    //                 isOrIncludes(
    //                   state.combined[root.type!][id]![field.foreign],
    //                   rootId,
    //                 );
    //               const prevIndex = rootRecordIds[rootId].indexOf(id);
    //               if (included && prevIndex === -1) {
    //                 addRecord(id);
    //               }
    //               if (!included && prevIndex !== -1) {
    //                 maybeRemoved[id] = true;
    //                 removeRecord(id);
    //               }
    //             });
    //           }
    //         }
    //       }
    //     }
    //     for (const rootId of rootChanges.added) {
    //       rootIds.push(rootId);
    //       initRootRecords(rootId);
    //     }
    //     const extraRemoved = Object.keys(maybeRemoved).filter(id =>
    //       rootIds.every(rootId => !rootRecordIds[rootId].includes(id)),
    //     );
    //     extraRemoved.forEach(id => delete records[id]);
    //     for (const id of Object.keys(changes[field.type] || {})) {
    //       if (records[id] && !newRecords.includes(id)) {
    //         for (const f of Object.keys(changes[field.type][id] || {})) {
    //           if (scalarFields[f]) {
    //             const value = ((state.combined[field.type] || {})[id] || {})[f];
    //             if (value === undefined) delete records[id][f];
    //             else records[id][f] = value;
    //           }
    //         }
    //       }
    //     }
    //     changesEmitter.emit({
    //       changes,
    //       rootChanges: {
    //         added: newRecords,
    //         removed: [...filteredRemoved, ...extraRemoved],
    //       },
    //     });
    //   });
    // return () => {
    //   stopRelations.forEach(s => s());
    //   stop && stop();
    // };
}
exports.default = readLayer;