"use strict";

Object.defineProperty(exports, "__esModule", { value: true });
var graphql_1 = require("graphql");
var language_1 = require("graphql/language");
var utils_1 = require("./utils");
var _parseLiteral = function _parseLiteral(ast, kinds) {
    if (ast.kind === language_1.Kind.STRING && (!kinds || kinds[language_1.Kind.STRING])) return ast.value;
    if (ast.kind === language_1.Kind.BOOLEAN && (!kinds || kinds[language_1.Kind.BOOLEAN])) return ast.value;
    if (ast.kind === language_1.Kind.INT && (!kinds || kinds[language_1.Kind.INT])) return parseInt(ast.value, 10);
    if (ast.kind === language_1.Kind.FLOAT && (!kinds || kinds[language_1.Kind.FLOAT])) return parseFloat(ast.value);
    if (ast.kind === language_1.Kind.OBJECT && (!kinds || kinds[language_1.Kind.OBJECT])) {
        return utils_1.keysToObject(ast.fields, function (f) {
            return _parseLiteral(f.value, kinds);
        }, function (f) {
            return f.name.value;
        });
    }
    if (ast.kind === language_1.Kind.LIST && (!kinds || kinds[language_1.Kind.LIST])) {
        return ast.values.map(function (v) {
            return _parseLiteral(v, kinds);
        });
    }
    return null;
};
var buildScalarTypes = function buildScalarTypes(types) {
    return utils_1.keysToObject(Object.keys(types), function (name) {
        return {
            decode: types[name].decode,
            encode: types[name].encode,
            type: new graphql_1.GraphQLScalarType({
                name: name,
                description: name + " custom scalar type",
                serialize: types[name].encode || function (value) {
                    return value;
                },
                parseValue: types[name].decode || function (value) {
                    return value;
                },
                parseLiteral: function parseLiteral(ast) {
                    return _parseLiteral(ast, types[name].kinds ? utils_1.keysToObject(types[name].kinds, function () {
                        return true;
                    }) : null);
                }
            })
        };
    });
};
var memoize = function memoize(func) {
    var cache = {};
    return function (value) {
        var key = JSON.stringify(value);
        if (cache[key] !== undefined) return cache[key];
        var result = func(value);
        cache[key] = result;
        return result;
    };
};
exports.default = Object.assign({ boolean: { type: graphql_1.GraphQLBoolean }, int: { type: graphql_1.GraphQLInt }, float: { type: graphql_1.GraphQLFloat }, string: { type: graphql_1.GraphQLString } }, buildScalarTypes({
    date: {
        encode: function encode(value) {
            return value ? new Date(value).getTime() : null;
        },
        decode: memoize(function (value) {
            return new Date(value);
        }),
        kinds: [language_1.Kind.INT, language_1.Kind.STRING]
    },
    file: {
        kinds: [language_1.Kind.STRING]
    },
    json: {}
}));