"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const resolveBatches = async (batchResolveFn, batches) => {
    for (const { sources, args, context, info, callbacks } of batches) {
        try {
            const values = await batchResolveFn(sources, args, context, info);
            callbacks.forEach(({ resolve, reject }, i) => {
                if (values[i] instanceof Error)
                    reject(values[i]);
                else
                    resolve(values[i]);
            });
        }
        catch (error) {
            callbacks.forEach(({ reject }) => reject(error));
        }
    }
};
class Batcher {
    constructor(batchResolveFn) {
        this.batches = new Map();
        this.hasScheduledResolve = false;
        this.batchResolveFn = batchResolveFn;
        this.batches = new Map();
    }
    add(source, args, context, info, resolve, reject) {
        const batchKey = `${context.rootQuery}:${JSON.stringify(info.fieldNodes)}`;
        let batch = this.batches.get(batchKey);
        if (!batch) {
            batch = { sources: [], args, context, info, callbacks: [] };
            this.batches.set(batchKey, batch);
        }
        batch.sources.push(source);
        batch.callbacks.push({ resolve, reject });
        if (!this.hasScheduledResolve) {
            this.hasScheduledResolve = true;
            process.nextTick(() => {
                const tickBatches = Array.from(this.batches.values());
                this.batches.clear();
                resolveBatches(this.batchResolveFn, tickBatches);
                this.hasScheduledResolve = false;
            });
        }
    }
}
function batchResolve(batchResolveFn) {
    const batcher = new Batcher(batchResolveFn);
    return (source, args, context, info) => new Promise((res, rej) => batcher.add(source, args, context, info, res, rej));
}
exports.default = batchResolve;
