"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
const core_1 = require("../core");
async function mutate(fields, connectors, runQuery, args, { user, mutationsInfo, }, auth) {
    const typeNames = Object.keys(args);
    const newIds = {};
    for (const type of typeNames) {
        newIds[type] = newIds[type] || {};
        args[type]
            .map(m => m.id)
            .filter(id => id[0] === '$')
            .forEach(id => {
            newIds[type][id] = connectors[type].newId();
        });
    }
    const getId = (type, id) => (Object.assign({}, (newIds[type] || {}), { $user: (user && user.id) || '' })[id] || id);
    const mutations = core_1.keysToObject(typeNames, () => []);
    for (const type of typeNames) {
        for (let _a of args[type]) {
            const { id } = _a, mutation = __rest(_a, ["id"]);
            const mId = getId(type, id);
            for (const f of Object.keys(fields[type])) {
                const field = fields[type][f];
                if (core_1.fieldIs.relation(field) && mutation[f]) {
                    mutation[f] = field.isList
                        ? mutation[f].map(v => getId(field.type, v))
                        : getId(field.type, mutation[f]);
                }
            }
            const data = Object.keys(mutation).length ? mutation : null;
            const prev = (id === mId && (await connectors[type].findById(mId))) || null;
            if (prev)
                delete prev.id;
            if (auth && type === auth.type && data && data[auth.usernameField]) {
                const { username, password } = JSON.parse(data[auth.usernameField]);
                delete data[auth.usernameField];
                if (!prev || prev[auth.usernameField] !== username) {
                    const existingUser = await connectors[type].query({
                        filter: [
                            'AND',
                            [[auth.usernameField, '=', username], ['id', '!=', mId]],
                        ],
                        end: 1,
                        fields: ['id'],
                    });
                    if (existingUser.length > 0) {
                        const error = new Error('Username already exists');
                        error.status = 400;
                        return error;
                    }
                }
                if (username)
                    data[auth.usernameField] = username;
                if (password && (!prev || !prev[auth.authIdField])) {
                    data[auth.authIdField] = password;
                }
            }
            const combinedData = Object.assign({}, prev, data);
            if (auth &&
                !await auth.allowMutation(fields, runQuery, user, type, mId, data && combinedData, prev)) {
                const error = new Error('Not authorized');
                error.status = 401;
                return error;
            }
            if (data) {
                for (const f of Object.keys(combinedData)) {
                    const field = fields[type][f];
                    if (!core_1.fieldIs.foreignRelation(field)) {
                        if (field.isList && data && data[f] && data[f].length === 0) {
                            data[f] = null;
                        }
                    }
                }
            }
            mutations[type].push({ id: mId, data, prev });
        }
    }
    const results = core_1.keysToObject(typeNames, () => []);
    for (const type of typeNames) {
        mutationsInfo.mutations[type] = mutationsInfo.mutations[type] || [];
        mutationsInfo.newIds[type] = Object.assign({}, mutationsInfo.newIds[type], newIds[type]);
        for (const { id, data, prev } of mutations[type]) {
            if (data) {
                const time = new Date();
                const fullData = Object.assign({}, !prev ? { createdat: time } : {}, { modifiedat: time }, data);
                if (auth && type === auth.type && fullData[auth.authIdField]) {
                    const username = fullData[auth.usernameField];
                    const password = fullData[auth.authIdField];
                    fullData[auth.authIdField] = await auth.createAuth(username, password, id, auth.metaFields &&
                        core_1.keysToObject(auth.metaFields.filter(f => fullData[f] !== undefined && fullData[f] !== null), f => fullData[f]));
                    mutationsInfo.newIds['$user'] = { username, password };
                }
                if (prev) {
                    console.log(`kalambo-mutate-update, ${type}:${id}, ` +
                        `old: ${JSON.stringify(prev)}, new: ${JSON.stringify(fullData)}`);
                    await connectors[type].update(id, fullData);
                }
                else {
                    console.log(`kalambo-mutate-insert, ${type}:${id}, new: ${JSON.stringify(fullData)}`);
                    await connectors[type].insert(id, fullData);
                }
                mutationsInfo.mutations[type].push({ id, data: fullData, prev });
                results[type].push(Object.assign({ id }, prev, fullData));
            }
            else {
                console.log(`kalambo-mutate-delete, ${type}:${id}, old: ${JSON.stringify(prev)}`);
                await connectors[type].delete(id);
                mutationsInfo.mutations[type].push({ id, data, prev });
                results[type].push({ id });
            }
        }
    }
    return results;
}
exports.default = mutate;
