"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const core_1 = require("../../core");
function memoryConnector(newId, initialData = []) {
    const records = initialData;
    return {
        newId,
        async query({ filter, sort, start = 0, end }) {
            if (start === end)
                return [];
            const filterFunc = (record) => core_1.runFilter(filter, record.id, record);
            const compareFunc = core_1.createCompare((record, key) => record[key], sort);
            return _.cloneDeep(records
                .filter(filterFunc)
                .sort(compareFunc)
                .slice(start, end));
        },
        async findById(id) {
            return _.cloneDeep(records.find(record => record.id === id) || null);
        },
        async insert(id, data) {
            records.push(Object.assign({ id }, data));
        },
        async update(id, data) {
            Object.assign(records.find(record => record.id === id), data);
        },
        async delete(id) {
            records.splice(records.findIndex(record => record.id === id), 1);
        },
        async dump() {
            return records;
        },
        async restore(data) {
            records.splice(0, records.length, data);
        },
    };
}
exports.default = memoryConnector;
