"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const graphql_1 = require("graphql");
const core_1 = require("../core");
function normalize(typeFields, data, queryDoc, argTypes, result) {
    const operationNode = queryDoc.definitions[0];
    if (operationNode.operation === 'mutation') {
        const rootNodes = operationNode.selectionSet.selections[0]
            .selectionSet.selections;
        for (const node of rootNodes) {
            const type = node.name.value;
            const fieldNodes = node.selectionSet.selections;
            const scalarFields = fieldNodes
                .filter(({ name, selectionSet }) => name.value !== 'id' && !selectionSet)
                .map(node => node.name.value);
            const relationFields = fieldNodes
                .filter(({ selectionSet }) => selectionSet)
                .map(node => node.name.value);
            data[type] = data[type] || {};
            result.commit[type].forEach(record => record &&
                (data[type][record.id] = Object.assign({}, (data[type][record.id] || {}), core_1.keysToObject(scalarFields, f => record[f]), core_1.keysToObject(relationFields, f => record[f] && core_1.mapArray(record[f], rec => rec && rec.id)))));
        }
        return {};
    }
    const firstIds = {};
    const processLayer = (root, field, { arguments: argNodes, selectionSet }, queryResults, path) => {
        const fieldNodes = selectionSet.selections;
        const scalarFields = fieldNodes
            .filter(({ name, selectionSet }) => name.value !== 'id' && !selectionSet)
            .map(node => node.name.value);
        const relationFields = fieldNodes
            .filter(({ selectionSet }) => selectionSet)
            .map(node => node.name.value);
        const args = core_1.keysToObject(argNodes || [], ({ value, name }) => graphql_1.valueFromAST(value, argTypes[name.value].type), ({ name }) => name.value);
        data[field.type] = data[field.type] || {};
        if (!root.type ||
            core_1.fieldIs.foreignRelation(field) ||
            (field.isList && args.sort)) {
            firstIds[path] = {};
        }
        Object.keys(queryResults).forEach(rootId => {
            if (root.type && core_1.fieldIs.relation(field) && field.isList && !args.sort) {
                if (data[root.type][rootId][root.field]) {
                    data[root.type][rootId][root.field].unshift(args.start || 0);
                }
            }
            queryResults[rootId].forEach((record, index) => record &&
                (!args.trace ||
                    args.trace.start === undefined ||
                    index < args.trace.start ||
                    args.trace.end === undefined ||
                    index >= args.trace.end) &&
                (data[field.type][record.id] = Object.assign({}, (data[field.type][record.id] || {}), core_1.keysToObject(scalarFields, f => record[f]), core_1.keysToObject(relationFields, f => record[f] && core_1.mapArray(record[f], rec => rec && rec.id)))));
            if (firstIds[path]) {
                firstIds[path][rootId] = (queryResults[rootId][args.offset || 0] || {}).id;
            }
        });
        fieldNodes.filter(({ selectionSet }) => selectionSet).forEach(node => processLayer({ type: field.type, field: node.name.value }, typeFields[field.type][node.name.value], node, Object.keys(queryResults).reduce((res, rootId) => (Object.assign({}, res, core_1.keysToObject(queryResults[rootId].filter(record => record), record => Array.isArray(record[node.name.value])
            ? record[node.name.value]
            : [record[node.name.value]], record => record.id))), {}), `${path}_${node.name.value}`));
    };
    const rootNodes = operationNode.selectionSet.selections;
    rootNodes.forEach(node => processLayer({ field: node.name.value }, { type: node.name.value, isList: true }, node, { '': result[node.name.value] || [] }, node.name.value));
    return { firstIds };
}
exports.default = normalize;
